// Page.cpp
//
// page & pageset objects

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <windows.h>
#include <assert.h>

#include "..\utility.h"

#include "page.h"

#define	RADIX_SIZE	27*27

// Page

Page::Page(const char* title, int flags)
{
	mTitle = new char[strlen(title) + 1];
	strcpy(mTitle, title);

	mASCII = NULL;
	mNext = NULL;
	mFlags = flags;
	mSeqNo = 0;
	memset(&mTime, 0, sizeof(mTime));

	mLink = NULL;
	mLinkCnt = 0;
}

Page::~Page()
{
	assert(!mASCII);
	assert(!mNext);
	delete[] mTitle;
}

// ModifyFlags
//
// change a page's flags

int Page::ModifyFlags(int on, int off)
{
	on &= ~PF_CHANGED;
	off &= ~PF_CHANGED;

	mFlags &= ~off;
	mFlags |= on;

	return mFlags;
}

// GetDateString
//
// get date as a (static) string

const char* Page::GetDateString() const
{
	static char	date[64];

	SYSTEMTIME	stime;

	GetTimeSys(&stime);

	sprintf(date, "%s %d, %d", MonthName(stime.wMonth), stime.wDay, stime.wYear);

	return date;
}

// GetTimeString
//
// get time as a (static) string

const char* Page::GetTimeString() const
{
	static char	time[128];

	SYSTEMTIME	stime;

	GetTimeSys(&stime);

	sprintf(time, "%s at %d:%d", GetDateString(), stime.wHour, stime.wMinute);

	return time;
}

// GetTimeSys
//
// get system time

void Page::GetTimeSys(SYSTEMTIME* stime) const
{
	FILETIME	ftime;

	FileTimeToLocalFileTime(&mTime, &ftime);

	FileTimeToSystemTime(&ftime, stime);
}

// PageSet

PageSet::PageSet(const char* pathname)
{
	mPathname = new char[strlen(pathname) + 1];
	strcpy(mPathname, pathname);

	mPageRadix = new Page*[RADIX_SIZE];
	memset(mPageRadix, 0, sizeof(Page*)*RADIX_SIZE);

	mPageCnt = 0;

	// load database
	char			fn[MAX_PATH];
	WIN32_FIND_DATA	fdata;

	sprintf(fn, "%s\\*.txt", pathname);

	HANDLE hnd = FindFirstFile(fn, &fdata);

	if (hnd != INVALID_HANDLE_VALUE)
	{
		char * nFileName = new char[MAX_PATH];
		do
		{
			fdata.cFileName[strlen(fdata.cFileName) - 4] = '\0';
			strcpy(nFileName, fdata.cFileName); //K+
			AsciiToUtf8(nFileName); //K+
//			Page* pp = CreatePage(fdata.cFileName);
			Page* pp = CreatePage(nFileName); //K+
			pp->mTime = fdata.ftLastWriteTime;

			LoadPage(pp);
			ScanForLinks(pp);
			UnloadPage(pp);

		} while (FindNextFile(hnd, &fdata));
		free (nFileName);
	}

	// now merge in dump file
	FILE*	fd = fopen("DumpAllPages", "rb");
	if (fd)
	{
		AddFromDump(fd);
		fclose(fd);
		DeleteFile("DumpAllPages.bak");
		MoveFile("DumpAllPages", "DumpAllPages.bak");
	}
}

PageSet::~PageSet()
{
	for (int ii = 0; ii < RADIX_SIZE; ii++)
	{
		Page*	pp = mPageRadix[ii];

		while (pp)
		{
			Page*	next = pp->mNext;
			UnloadPage(pp, true);
			pp->mNext = NULL;
			delete pp;
			pp = next;
		}
	}

	delete[] mPathname;
}

static int encode(char c)
{
	if ((c >= 'a') && (c <= 'z'))	return c - 'a';
	if ((c >= 'A') && (c <= 'Z'))	return c - 'A';
	return 26;
}

int	PageSet::MakeRadix(const char* title)
{
	assert(RADIX_SIZE == 27*27);

	if (title[0])
	{
		return encode(title[0]) + 27 * encode(title[1]);
	}

	return RADIX_SIZE - 1;
}

Page* PageSet::FindPage(const char* title) const
{
//K	if (!IsValidTitle(title))	return NULL;

	int radix = MakeRadix(title);

	Page*	pp;

	for (pp = mPageRadix[radix]; pp; pp = pp->mNext)
	{
		if (!stricmp(title, pp->mTitle))	return pp;
	}

	return NULL;
}

Page* PageSet::CreatePage(const char* title, int flags)
{
//K	if (!IsValidTitle(title))	return NULL;
	Page*	pp = FindPage(title);

	if (!pp)
	{
		int	radix = MakeRadix(title);

		pp = new Page(title, flags);
		pp->mNext = mPageRadix[radix];
		mPageRadix[radix] = pp;

		mPageCnt++;
	}

	return pp;
}

bool PageSet::LoadPage(Page* page) const
{
	if (page->mASCII)			return true;

//	char	fn[MAX_PATH];
	char* fn;
	char* aTitle = new char[strlen(page->mTitle)+1];	//Z+

	strcpy(aTitle, page->mTitle);	//Z+
	Utf8ToAscii(aTitle);			//Z+
	fn = new char[strlen(mPathname)+strlen(aTitle)+6];	//Z+

	sprintf(fn, "%s\\%s.txt", mPathname, aTitle);

//	char * afn = new char [MAX_PATH];			//K+
//	strcpy(afn, fn);	//K+
//	Utf8ToAscii(afn);	//K+

	FILE*	fd = fopen(fn, "rb");
//	FILE*	fd = fopen(afn, "rb");
	if (!fd)	return false;

	fseek(fd, 0, SEEK_END);
	size_t len = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	page->mASCII = new char[len + 1];
	fread(page->mASCII, 1, len, fd);
	fclose(fd);
	page->mASCII[len] = 0;

//	free(afn);
	delete[] fn;

	return true;
}

bool PageSet::UnloadPage(Page* page, bool force) const
{
	bool	rc = true;

	if (!page->mASCII)			return rc;

	if (page->mFlags & PF_CHANGED)
	{
		if (!WritePage(page) && !force)	return false;
	}

	if ((page->mFlags & PF_KEEP) && !force)	return false;

	delete[] page->mASCII;
	page->mASCII = NULL;

	return rc;
}

bool PageSet::WritePage(Page* page) const
{
	assert(page->mASCII);

	char* fn;
	char* aTitle = new char[strlen(page->mTitle)+1];	//Z+

	strcpy(aTitle, page->mTitle);	//Z+
	Utf8ToAscii(aTitle);			//Z+
	fn = new char[strlen(mPathname)+strlen(aTitle)+6];	//Z+

	sprintf(fn, "%s\\%s.txt", mPathname, aTitle);

//	char * afn = new char[MAX_PATH];
//	strcpy(afn, fn);
//	Utf8ToAscii(afn);	//K+

	FILE*	fd = fopen(fn, "wb");
//	FILE*	fd = fopen(afn, "wb");
	if (!fd)	return false;

	page->mFlags &= ~PF_CHANGED;
	fwrite(page->mASCII, 1, strlen(page->mASCII), fd);
	fclose(fd);

	WIN32_FIND_DATA	fdata;
	HANDLE	hf = FindFirstFile(fn, &fdata);
//	HANDLE	hf = FindFirstFile(afn, &fdata);
	if (hf != INVALID_HANDLE_VALUE)
	{
		page->mTime = fdata.ftLastWriteTime;
		FindClose(hf);
	}

	ScanForLinks(page);

	delete[] fn;	//K+

	return true;
}

void PageSet::EditPage(Page* page, const char* src, size_t srccnt) const
{
	delete[] page->mASCII;
	page->mASCII = new char[srccnt + 1];
	memcpy(page->mASCII, src, srccnt);
	page->mASCII[srccnt] = 0;

	page->mSeqNo++;
	page->mFlags |= PF_CHANGED;

	UnloadPage(page);
}

bool PageSet::WriteEditPage(const char* pname, int nrows, int ncols, FILE* fd) const
{
	Page*	page = FindPage(pname);
//K	if (!page && !IsValidTitle(pname))	return false;

	if (page && (page->GetFlags() & PF_META))	return false;

	char*	spname = PageSet::MakeSpacedTitle(pname);

	nrows = nrows ? nrows : 18;
	ncols = ncols ? ncols : 80;

	fprintf(fd, "<html><head>%s<title>Edit %s</title></head>\n", METATAG, spname);
	fprintf(fd, "<body><form method=\"POST\" action=\"/\">\n");
	fprintf(fd, "<h1><a href=\"StartPage\"><img src=\"%s\"></a> ", gIconName);
	fprintf(fd, "%s\n", spname);
	fprintf(fd, "<input type=\"submit\" value=\" Save \">\n");
	fprintf(fd, "<input type=\"reset\" value=\" Reset \"></h1>\n");
	fprintf(fd, "<p><blockquote><textarea name=\"text\" rows=%d cols=%d wrap=virtual>\n", nrows, ncols);
	if (page)
	{
		WriteAsTextArea(page, fd);
	}
	else
	{
		fprintf(fd, "Define %s here.\n", pname);
	}
	fprintf(fd, "</textarea></blockquote><p>\n");
	fprintf(fd, "<a href=\"EditPage?page=%s", pname);
	if (nrows > 20)
	{
		fprintf(fd, "&rows=18&cols=80\">EditPage</a> using a smaller text area.<br>\n");
	}
	else
	{
		fprintf(fd, "&rows=34&cols=96\">EditPage</a> using a larger text area.<br>\n");
	}
	fprintf(fd, "<input type=\"hidden\" name=\"page\" value=\"%s\">\n", pname);
	fprintf(fd, "<input type=\"hidden\" name=\"seqno\" value=\"%d\">\n", page ? page->GetSeqNo() : 0);
	fprintf(fd, "</form></body></html>\n");

	delete[] spname;

	return true;
}

bool PageSet::WriteSearchForm(const char* name, const char* value, FILE *fd) const
{
	fprintf(fd, "<form method=\"POST\" action=\"/\">\n");
	fprintf(fd, "<input type=\"text\" size=\"40\" name=\"%s\" value=\"%s\">\n", name, value);
	fprintf(fd, "<input type=\"submit\" value=\" Go \">\n");
	fprintf(fd, "</form>\n");

	return true;
}

bool PageSet::WriteMergePage(Page* page, const char* newv, FILE* fd) const
{
	const char*	pname = page->GetTitle();
	char*	spname = PageSet::MakeSpacedTitle(pname);

	fprintf(fd, "<html><head>%s<title>Merge %s</title></head>\n", METATAG, spname);
	fprintf(fd, "<body><form method=\"POST\" action=\"/\">\n", pname);
	fprintf(fd, "<h1><a href=\"StartPage\"><img src=\"%s\"></a> ", gIconName);
	fprintf(fd, "%s\n", spname);
	fprintf(fd, "<input type=\"submit\" value=\" Save \">\n");
	fprintf(fd, "<input type=\"reset\" value=\" Reset \"></h1>\n");
	fprintf(fd, "<p>Another user edited the page while you were working.\n");
	fprintf(fd, "<p>The current page is shown in the top window; the page you submitted is shown in the bottom window.\n");
	fprintf(fd, "<p>Please merge your changes from the bottom window into the top window, and resave.\n");
	fprintf(fd, "<p><blockquote><textarea name=\"text\" rows=%d cols=%d wrap=virtual>\n", 18, 80);
	WriteAsTextArea(page, fd);
	fprintf(fd, "</textarea><p><textarea name=\"newv\" rows=%d cols=%d wrap=virtual>\n", 18, 80);
	WriteAsTextArea(newv, fd);
	fprintf(fd, "</textarea></blockquote><p>\n");
	fprintf(fd, "<input type=\"hidden\" name=\"page\" value=\"%s\">\n", pname);
	fprintf(fd, "<input type=\"hidden\" name=\"seqno\" value=\"%d\">\n", page->GetSeqNo());
	fprintf(fd, "</form></body></html>\n");

	delete[] spname;

	return true;
}

bool PageSet::WriteEditedMessage(Page* page, FILE* outfd) const
{
	const char*	title = page->GetTitle();

	fprintf(outfd, "<html><head>%s<title>Thank you for editing %s</title></head>\n", METATAG, title);
	fprintf(outfd, "<body><form><h1>Thank you for editing <a href=\"%s\">%s</a></h1><p>\n", title, title);
	fprintf(outfd, "</html>\n");

	return true;
}

bool PageSet::WriteAsTextArea(const char* text, FILE* outfd) const
{
	if (text)
	{
		while (char ch = *text++)
		{
			if (ch == '&')			fprintf(outfd, "&amp;");
			else if (ch == '\"')	fprintf(outfd, "&quot;");
			else if	(ch == '<')		fprintf(outfd, "&lt;");
			else if (ch == '>')		fprintf(outfd, "&gt;");
			else if (ch == 0x0D)	;
			else					fputc(ch, outfd);
		}
	}

	return true;
}

bool PageSet::WriteAsTextArea(Page* page, FILE* outfd) const
{
	LoadPage(page);

	bool rc = WriteAsTextArea(page->mASCII, outfd);

	UnloadPage(page);

	return rc;
}

bool PageSet::IsValidTitle(const char* tptr)
{
	// Must be <cap><small>
	if ((*tptr < 'A') || (*tptr > 'Z'))	return false;
	tptr++;
	if ((*tptr < 'a') || (*tptr > 'z'))	return false;
	tptr++;

	// Skip past all <small>
	while (*tptr)
	{
		if ((*tptr >= 'a') && (*tptr <= 'z'))	tptr++;
		else									break;
	}

	// Now <cap> or <numeral>
	if (((*tptr < 'A') || (*tptr > 'Z')) &&
		((*tptr < '0') || (*tptr > '9')))
	{
		return false;
	}

	// And no invalid rubbish afterwards
	while (*++tptr)
	{
		if ((*tptr >= 'A') && (*tptr <= 'Z'))	continue;
		if ((*tptr >= 'a') && (*tptr <= 'z'))	continue;
		if ((*tptr >= '0') && (*tptr <= '9'))	continue;
		return false;
	}

	// It's OK
	return true;
}

char* PageSet::MakeSpacedTitle(const char* title)
{
//K+
	if (!IsValidTitle(title)) {
		char*	out = new char[strlen(title) + 1];
		strcpy(out, title);
		return out;
	}

	const char*	tptr = title;
	int			caps = 0;
	
	while (*tptr)
	{
		if ((*tptr >= 'A') && (*tptr <= 'Z'))	caps++;
		tptr++;
	}
	caps--;	// ignore first

	char*	obuf = new char[strlen(title) + caps + 1];
	char*	optr = obuf;
	tptr = title;
	*optr++ = *tptr++;
	while (*tptr)
	{
		if ((*tptr >= 'A') && (*tptr <= 'Z'))	*optr++ = ' ';	// but I think I prefer no space ;)
		*optr++ = *tptr++;
	}
	*optr++ = 0;

	return obuf;
}

Page**	PageSet::MakeSortedList(int (*compare)(const Page** pp1, const Page** pp2)) const
{
	Page**	array = new Page*[mPageCnt];
	Page**	aptr = array;

	for (int ii = 0; ii < RADIX_SIZE; ii++)
	{
		for (Page* pp = mPageRadix[ii]; pp; pp = pp->mNext)
		{
			*aptr++ = pp;
		}
	}

	qsort(array, mPageCnt, sizeof(Page*), (int (*)(const void*, const void*))compare);

	return array;
}

static int AlphaCompare(const Page** pp1, const Page** pp2)
{
	return stricmp((*pp1)->GetTitle(), (*pp2)->GetTitle());
}

Page**	PageSet::MakeAlphabeticList() const
{
	return MakeSortedList(AlphaCompare);
}

static int TimeCompare(const Page** pp1, const Page** pp2)
{
	FILETIME	t1;
	FILETIME	t2;

	t1 = (*pp1)->GetTime();
	t2 = (*pp2)->GetTime();

	if (t1.dwHighDateTime < t2.dwHighDateTime)	return +1;
	if (t1.dwHighDateTime > t2.dwHighDateTime)	return -1;

	if (t1.dwLowDateTime < t2.dwLowDateTime)	return +1;
	if (t1.dwLowDateTime > t2.dwLowDateTime)	return -1;

	return 0;
}

Page**	PageSet::MakeRecentList() const
{
	return MakeSortedList(TimeCompare);
}

void PageSet::ListAllPages(FILE* fd) const
{
	Page**	list = MakeAlphabeticList();

	fprintf(fd, "<html><head>%s<title>ListAllPages</title></head>\n", METATAG);
	fprintf(fd, "<body><h1><a href=\"StartPage\"><img src=\"%s\"></a> List All Pages</h1><p>\n", gIconName);

	fprintf(fd, "The following is a complete list of pages on this server:<p>\n<ul>\n");

	for (int ii = 0; ii < mPageCnt; ii++)
	{
		if (list[ii]->GetFlags() & PF_META)	continue;

		fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", list[ii]->GetTitle(), list[ii]->GetTitle());
	}

	fprintf(fd, "</ul><hr></body></html>\n");

	delete[] list;
}

void PageSet::RecentChanges(FILE* fd) const
{
	Page**	list = MakeRecentList();

	fprintf(fd, "<html><head>%s<title>Recent Changes</title></head>\n", METATAG);
	fprintf(fd, "<body><h1><a href=\"StartPage\"><img src=\"%s\"></a> Recent Changes</h1><p>\n", gIconName);

	fprintf(fd, "The following pages have recently been edited:<p>\n<ul>\n");

	SYSTEMTIME	ctime;
	memset(&ctime, 0, sizeof(ctime));

	int	ndays = 7;
	
	for (int ii = 0; ii < mPageCnt; ii++)
	{
		FILETIME	ftime = list[ii]->GetTime();

		if (!ftime.dwHighDateTime && !ftime.dwLowDateTime)	continue;

		SYSTEMTIME	stime;

		list[ii]->GetTimeSys(&stime);

		if ((stime.wYear != ctime.wYear) ||
			(stime.wMonth != ctime.wMonth) ||
			(stime.wDay != ctime.wDay))
		{
			ndays--;
			if (!ndays)	break;

			fprintf(fd, "</ul><b>%s</b><p><ul>\n", list[ii]->GetDateString());
			ctime = stime;
		}
		fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", list[ii]->GetTitle(), list[ii]->GetTitle());
	}
	fprintf(fd, "</ul><hr></body></html>\n");

	delete[] list;
}

static const char hdr[13] = "WikiFileDump";

void PageSet::DumpAllPages(FILE* fd) const
{
	fwrite(hdr, 1, 12, fd);

	size_t	len;

	for (int ii = 0; ii < RADIX_SIZE; ii++)
	{
		for (Page* pp = mPageRadix[ii]; pp; pp = pp->mNext)
		{
			if (pp->GetFlags() & PF_META)	continue;

			LoadPage(pp);

			len = strlen(pp->GetTitle());
			fwrite(&len, 1, sizeof(size_t), fd);
			fwrite(pp->GetTitle(), 1, len, fd);

			len = strlen(pp->mASCII);
			fwrite(&len, 1, sizeof(size_t), fd);
			fwrite(pp->mASCII, 1, len, fd);

			UnloadPage(pp);
		}
	}

	len = 0;
	fwrite(&len, 1, sizeof(size_t), fd);
}

void PageSet::AddFromDump(FILE* fd)
{
	char fhdr[12];
	fread(fhdr, 1, 12, fd);

	if (memcmp(hdr, fhdr, 12))	return;	// invalid header

	for (;;)
	{
		size_t	len;

		// read title
		fread(&len, 1, sizeof(size_t), fd);
		if (!len)	break;

		char*	name = new char[len + 1];
		fread(name, 1, len, fd);
		name[len] = 0;

		// create page
		Page*	pp = CreatePage(name);
		delete[] name;

		// read data
		fread(&len, 1, sizeof(size_t), fd);
		char*	data = new char[len + 1];
		fread(data, 1, len, fd);
		data[len] = 0;

		EditPage(pp, data, len);
	}
}

// ScanForLinks
//
// find all the links in a document
//
// returns a NULL-terminated list

void PageSet::ScanForLinks(Page* page) const
{
	if (!page->mASCII)	return;

	delete[] page->mLink;
	page->mLinkCnt = 0;

	char**	array = new char*[32];
	int		ix = 0;
	int		max = 32;

	const char*	doc = page->mASCII;

	while (*doc)
	{
		if (isalpha(*doc))
		{
			char*	word = GetAlphaNumeric(doc);

			if (IsValidTitle(word))
			{
				// is it already in the array?
				for (int ii = 0; ii < ix; ii++)
				{
					if (!stricmp(array[ii], word))	break;
				}
				if (ii == ix)
				{
					// no!
					if (ix == max)
					{
						// extend array
						max *= 2;
						char**	narray = new char*[max];
						for (int ii = 0; ii < ix; ii++)
						{
							narray[ii] = array[ii];
						}
						delete[] array;
						array = narray;
					}
					array[ix++] = word;
				}
			}
			else
			{
				delete[] word;
			}
		}

//K+
		else if (*doc=='[') //Find ExtendedWikiName
		{
			char*	word = GetSquare(doc);

			if (word != NULL)
			{
				// is it already in the array?
				for (int ii = 0; ii < ix; ii++)
				{
					if (!stricmp(array[ii], word))	break;
				}
				if (ii == ix)
				{
					// no!
					if (ix == max)
					{
						// extend array
						max *= 2;
						char**	narray = new char*[max];
						for (int ii = 0; ii < ix; ii++)
						{
							narray[ii] = array[ii];
						}
						delete[] array;
						array = narray;
					}
					array[ix++] = word;
				}
			}
		}

		else
		{
			doc++;
		}
	}

	page->mLink = array;
	page->mLinkCnt = ix;
}

void PageSet::ReverseLink(Page* page, FILE* fd) const
{
	// make list of pages
	Page**	list = MakeAlphabeticList();

	// for each page, look for this title in its list
	for (int ii = 0; ii < mPageCnt; ii++)
	{
		Page*	pp = list[ii];

		for (int jj = 0; jj < pp->mLinkCnt; jj++)
		{
			if (!stricmp(pp->mLink[jj], page->mTitle))	break;
		}
		if (jj == pp->mLinkCnt)
		{
			// no link - clear list element
			list[ii] = NULL;
		}
	}

	// print HTML
	const char*	title = page->GetTitle();
	char*		stitle = MakeSpacedTitle(title);

	fprintf(fd, "<html>%s<head><title>ReverseLink - %s</title></head>\n", METATAG, stitle);
	fprintf(fd, "<body><h1><a href=\"StartPage\"><img src=\"%s\"></a> ReverseLink - <a href=\"%s\">%s<a></h1><p>\n", gIconName, title, stitle);

	delete[] stitle;

	fprintf(fd, "The following pages link to <a href=\"%s\">%s</a>:<p>\n<ul>\n", title, title);

	bool	any = false;

	for (ii = 0; ii < mPageCnt; ii++)
	{
		if (!list[ii])						continue;
		if (list[ii]->GetFlags() & PF_META)	continue;

		fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", list[ii]->GetTitle(), list[ii]->GetTitle());

		any = true;
	}

	fprintf(fd, "</ul>\n");
	if (!any)	fprintf(fd, "<blockquote><i>(None)</i></blockquote>\n");
	fprintf(fd, "<hr></body></html>\n");

	delete[] list;
}

void PageSet::TitleSearch(const char* search, FILE* fd) const
{
	// make list of pages
	Page**	list = MakeAlphabeticList();

	fprintf(fd, "<html>%s<head><title>Title Search - %s</title></head>\n", METATAG, search);
	fprintf(fd, "<body><h1><a href=\"StartPage\"><img src=\"%s\"></a> <a href=\"TitleSearch\">TitleSearch</a> - %s</h1><p>\n", gIconName, search);

	fprintf(fd, "The following pages were found:<p>\n<ul>\n");

	// for each page, look for the search string
	bool	any = false;

	for (int ii = 0; ii < mPageCnt; ii++)
	{
		Page*	pp = list[ii];
		
		if (pp->GetFlags() & PF_META)	continue;

		if (MatchWildcardString(pp->GetTitle(), search))
		{
			fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", pp->GetTitle(), pp->GetTitle());
			any = true;
		}
	}

	fprintf(fd, "</ul>\n");
	if (!any)	fprintf(fd, "<blockquote><i>(None)</i></blockquote>\n");
	fprintf(fd, "<hr></body></html>\n");

	delete[] list;
}

void PageSet::FullSearch(const char* search, FILE* fd) const
{
	// make list of pages
	Page**	list = MakeAlphabeticList();

	fprintf(fd, "<html><head>%s<title>Full Search - %s</title></head>\n", METATAG, search);
	fprintf(fd, "<body><h1><a href=\"StartPage\"><img src=\"%s\"></a> <a href=\"FullSearch\">FullSearch</a> - %s</h1><p>\n", gIconName, search);

	fprintf(fd, "The following pages were found:<p>\n<ul>\n");

	// for each page, load it in and look for the search string
	bool	any = false;

	for (int ii = 0; ii < mPageCnt; ii++)
	{
		Page*	pp = list[ii];

		if (pp->GetFlags() & PF_META)	continue;

		LoadPage(pp);

		if (MatchWildcardString(pp->mASCII, search))
		{
			fprintf(fd, "<li><a href=\"%s\">%s</a></li>\n", pp->GetTitle(), pp->GetTitle());
			any = true;
		}

		UnloadPage(pp);
	}		

	fprintf(fd, "</ul>\n");
	if (!any)	fprintf(fd, "<blockquote><i>(None)</i></blockquote>\n");
	fprintf(fd, "<hr></body></html>\n");
}
