// Page.h
//
// page & pageset objects

const char gIconName[] = "Icon.jpg";

class Page
{
public:
	Page(const char* title, int flags = 0);
	~Page();

	const char*	GetTitle() const	{ return mTitle; }
	int			GetFlags() const	{ return mFlags; }
	int			GetSeqNo() const	{ return mSeqNo; }
	FILETIME	GetTime() const		{ return mTime; }
	const char*	GetTimeString() const;	// get string with date & time
	const char*	GetDateString() const;	// get string with date
	void		GetTimeSys(SYSTEMTIME* stime) const;
	int			ModifyFlags(int on, int off);

private:

	friend class PageSet;

	char*		mTitle;		// title of page
	char*		mASCII;		// page contents in ASCII
	Page*		mNext;		// next page in radix
	int			mFlags;		// flags
	int			mSeqNo;		// sequence number
	FILETIME	mTime;		// filetime

	char**		mLink;		// list of links
	int			mLinkCnt;	// number of links
};

enum PageFlags
{
	PF_KEEP = 1,		// page is kept in RAM
	PF_CHANGED = 2,		// page differs from file version
	PF_META = 4			// page is a meta-page
};

struct LineFmt;

class PageSet
{
public:
	PageSet(const char* pathname);
	~PageSet();

	// basic manipulation
	Page*	FindPage(const char* title) const;
	Page*	CreatePage(const char* title, int flags = 0);

	// parse and write to file
	bool	WriteAsHTML(Page* page, FILE* fd) const;
	bool	WriteEditPage(const char* pname, int nrows, int ncols, FILE* fd) const;
	bool	WriteMergePage(Page* page, const char* newv, FILE* fd) const;
	bool	WriteEditedMessage(Page* page, FILE* fd) const;
	bool	WriteSearchForm(const char* name, const char* value, FILE* fd) const;

	// write as HTML-safe raw text
	bool	WriteAsTextArea(const char* text, FILE* fd) const;
	bool	WriteAsTextArea(Page* page, FILE* fd) const;

	// write to/from disk
	bool	LoadPage(Page* page) const;
	bool	UnloadPage(Page* page, bool force = false) const;
	bool	WritePage(Page* page) const;

	// change contents
	void	EditPage(Page* page, const char* src, size_t srccnt) const;

	// is a title valid?
	static bool	IsValidTitle(const char* candidate);

	// make a spaced title
	static char* MakeSpacedTitle(const char* title);

	// get pathname
	const char* GetPathName() const	{ return mPathname; }

	// whole database operations
	void	DumpAllPages(FILE* fd) const;
	void	AddFromDump(FILE* fd);

	void	ListAllPages(FILE* fd) const;
	void	RecentChanges(FILE* fd) const;
	void	ReverseLink(Page* page, FILE* fd) const;

	void	TitleSearch(const char* search, FILE* fd) const;
	void	FullSearch(const char* search, FILE* fd) const;

	// make a sorted list
	Page**	MakeSortedList(int (*compare)(const Page** pp1, const Page** pp2)) const;
	Page**	MakeAlphabeticList() const;
	Page**	MakeRecentList() const;

private:
	Page**	mPageRadix;
	char*	mPathname;
	int		mPageCnt;

	static int MakeRadix(const char* title);

	static char*	ReplacePostfix(const char* lp);

	void	ScanForLinks(Page* page) const;

	void	AcceptQuote(const char*& lp, FILE* fd, LineFmt& fmt) const;
	void	AcceptAlpha(const char*& lp, FILE* fd, LineFmt& fmt) const;
	void	AcceptExtended(const char*& lp, FILE* fd, LineFmt& fmt) const;
	void	AcceptBrace(const char*& lp, FILE* fd, LineFmt& fmt) const;
	void	WriteLine(const char* lp, FILE* fd, bool italic) const;
};

//K+
#define METATAG "<META http-equiv='Content-Type' content='text/html;charset=UTF-8'>"
