// PageFormatting.cpp
//
// formatting stuff for Page class

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <windows.h>
#include <assert.h>

#include "..\utility.h"

#include "page.h"

// LineFmt
//
// line formatting data

struct LineFmt
{
	bool	italic;
	bool	bold;
};

// ParaFmt
//
// paragraph formatting data

struct ParaFmt
{
	int		indent;		// amount of indent, 0 to 3
	char	indents[4];	// indent data; o = ol; u = ul
	int		quote;		// quote depth
	bool	pre;		// preformatted flag

	void	SetFmt(const ParaFmt& fmt, FILE* fd);
};

// AcceptQuote
//
// accept part of text beginning '

void PageSet::AcceptQuote(const char*& lp, FILE* fd, LineFmt& fmt) const
{
	int	quotes = 0;
	while (*lp == '\'')
	{
		quotes++;
		lp++;
	}
	if (quotes == 1)
	{
		fputc('\'', fd);
	}
	else
	{
		if (quotes != 3)	// 2 or >3 switches italic
		{
			fprintf(fd, fmt.italic ? "</i>" : "<i>");
			fmt.italic = !fmt.italic;
		}
		if (quotes != 2)	// 3 or >3 switches bold
		{
			fprintf(fd, fmt.bold ? "</b>" : "<b>");
			fmt.bold = !fmt.bold;
		}
	}
}

// AcceptAlpha
//
// accept part of text beginning with a letter

void PageSet::AcceptAlpha(const char*& lp, FILE* fd, LineFmt& fmt) const
{
	char*	word = GetAlphaNumeric(lp);

	if (IsValidTitle(word))
	{
		// Page link
		Page*	pp = FindPage(word);
		if (pp)
		{
			// link
			fprintf(fd, "<a href=\"%s\">%s</a>", word, word);
		}
		else
		{
			// dangling link
			fprintf(fd, "%s<a href=\"%s\">?</a>", word, word);
		}
	}
	else if (lp[0] == ':')
	{
		// URL
		lp -= strlen(word);		// back up to start of URL
		delete[] word;
		word = GetURL(lp);

		if (lp[-1] == ':')
		{
			// not a URL
			fprintf(fd, "%s", word);
		}
		else
		{
			// is a URL - check protocol
			if (!memcmp(word, "local://", 8))
			{
				// local://file.xxx is replaced with file.xxx
				fprintf(fd, "<a href=\"%s\">%s</a>", word + 8, word + 8);
			}
			else if (!memcmp(word, "ilocal://", 9))
			{
				// ilocal://file.xx is replaced with inline image
				fprintf(fd, "<img src=\"%s\">", word + 9, word + 9);
			}
			else if (!memcmp(word, "mailto:", 7))
			{
				fprintf(fd, "<a href=\"%s\">%s</a>", word, word + 7);
			}
			else
			{
				fprintf(fd, "<a href=\"%s\">%s</a>", word, word);
			}
		}
	}
	else if (!strcmp(word, "ISBN"))	// case sensitive
	{
		// ISBN 123-456-789X?
		char	rawisbn[11];

		char*	isbn = GetISBN(lp, rawisbn);

		if (isbn)
		{
			fprintf(fd, "<a href=\"http://www.amazon.com/exec/obidos/ISBN=%s\">", rawisbn);
			fprintf(fd, "ISBN%s</a>", isbn);
			delete[] isbn;
		}
		else
		{
			fprintf(fd, "%s", word);
		}
	}
	else
	{
		fprintf(fd, "%s", word);
	}

	delete[] word;
}


//K+
// AcceptExtended
//
// accept ExtendedWikiName
// accept part of text beginning with a letter

void PageSet::AcceptExtended(const char*& lp, FILE* fd, LineFmt& fmt) const
{

	// find [word]
	char* word = GetSquare(lp);
	
	if (word==NULL) {
		fprintf(fd, "[");
		return;
	}

	// Page link
	char* eword = new char[strlen(word)+1];
	strcpy (eword, word);
	//Utf8ToAscii(eword);
	URLEncode(eword);

	Page*	pp = FindPage(word);
	if (pp)
	{
		// link
		fprintf(fd, "<a href=\"%s\">%s</a>", eword, word);
	}
	else
	{
		// dangling link
		fprintf(fd, "%s<a href=\"%s\">?</a>", word, eword);
	}
	delete eword;
}

// AcceptBrace
//
//K accept part of text beginning [
// accept part of text beginning {

void PageSet::AcceptBrace(const char*& lp, FILE* fd, LineFmt& fmt) const
{
	lp++;

	char*	num = GetNumber(lp);

//K	if (num[0] && (*lp == ']'))
	if (num[0] && (*lp == '}'))
	{
		lp++;
//K		fprintf(fd, "<a href=\"#%s\">[%s]</a>", num, num);
		fprintf(fd, "<a href=\"#%s\">{%s}</a>", num, num);
	}
	else if (num[0])
	{
//K		fprintf(fd, "[%s", num);
		fprintf(fd, "{%s", num);
	}
	else
	{
		char*	word = GetAlphaNumeric(lp);

//K		if (*lp != ']')
		if (*lp != '}')
		{
//K			fprintf(fd, "[");
			fprintf(fd, "{");
			lp -= strlen(word);
			return;
		}

		if (!stricmp(word, "titlesearch"))
		{
			// title search
			WriteSearchForm("title_search", "", fd);
			lp++;
		}
		else if (!stricmp(word, "fullsearch"))
		{
			// full search
			WriteSearchForm("full_search", "", fd);
			lp++;
		}
		else
		{
//K			fprintf(fd, "[");
			fprintf(fd, "{");
			lp -= strlen(word);
		}
	}
}

// ReplacePostfix
//
// replace postfix commands with prefix commands

char* PageSet::ReplacePostfix(const char* lp)
{
	char*	nline = new char[strlen(lp) + 1];
	char*	nptr = nline;

	while (*lp)
	{
		int	count = 0;

		while ((lp[0] == '^') && (lp[1] == 'H'))
		{
			lp += 2;
			count++;
		}

		if (count)
		{
			if (nptr - nline < count)	count = nptr - nline;

			char*	tmp = new char[count];
			nptr -= count;
			memcpy(tmp, nptr, count);
			*nptr++ = 0x01;	// strike on
			memcpy(nptr, tmp, count);
			nptr += count;
			*nptr++ = 0x02;	// strike off
			delete[] tmp;
		}
		else
		{
			*nptr++ = *lp++;
		}
	}
	*nptr++ = 0;
	return nline;
}

// WriteLine
//
// write a line of text out

void PageSet::WriteLine(const char* lp, FILE* fd, bool italic) const
{
	LineFmt	fmt;

	fmt.italic = italic;
	fmt.bold = false;

	if (fmt.italic)	fprintf(fd, "<i>");

	// replace postfix commands
	char* nline = ReplacePostfix(lp);

	lp = nline;

	while (char ch = *lp)
	{
		if (isalpha(ch))
		{
			AcceptAlpha(lp, fd, fmt);
		}
		else
		{
			switch (ch)
			{
			case '\'':
				AcceptQuote(lp, fd, fmt);
				break;

//K			case '[':
			case '{':
				AcceptBrace(lp, fd, fmt);
				break;
//K+
			case '[':
				AcceptExtended(lp, fd, fmt);
				break;

			case '&':
				fprintf(fd, "&amp;");
				lp++;
				break;

			case '\"':
				fprintf(fd, "&quot;");
				lp++;
				break;

			case '<':
				fprintf(fd, "&lt;");
				lp++;
				break;

			case '>':
				fprintf(fd, "&gt;");
				lp++;
				break;

			case 0x0D:
				lp++;
				break;

			case 0x01:	// strike on
				fprintf(fd, "<strike>");
				lp++;
				break;

			case 0x02:	// strike off
				fprintf(fd, "</strike>");
				lp++;
				break;

			default:
				fputc(ch, fd);
				lp++;
				break;
			}
		}
	}

	delete[] nline;

	if (fmt.bold)		fprintf(fd, "</b>");
	if (fmt.italic)		fprintf(fd, "</i>");
}

void ParaFmt::SetFmt(const ParaFmt& fmt, FILE* fd)
{
	// turn block-quotes off
	while (fmt.quote < quote)
	{
		fprintf(fd, "</blockquote>\n");
		quote--;
	}
	// turn preformatting off
	if (pre && !fmt.pre)
	{
		fprintf(fd, "</pre>\n");
		pre = false;
	}

	// reduce indent
	while (fmt.indent < indent)
	{
		indent--;
		if (indents[indent] == 'o')	fprintf(fd, "</ol>\n");
		else						fprintf(fd, "</ul>\n");
	}

	// increase indent
	while (fmt.indent > indent)
	{
		indents[indent] = fmt.indents[indent];
		if (indents[indent] == 'o')	fprintf(fd, "<ol>\n");
		else						fprintf(fd, "<ul>\n");
		indent++;
	}

	// turn preformatting on
	if (fmt.pre && !pre)
	{
		fprintf(fd, "<pre>\n");
		pre = true;
	}
	// turn block-quotes on
	while (fmt.quote > quote)
	{
		fprintf(fd, "<blockquote>\n"); 
		quote++;
	}
}

static const char* SetLineAttributes(const char* line, ParaFmt& _fmt, FILE* fd)
{
	ParaFmt	fmt;

	fmt.indent = 0;
	fmt.quote = 0;
	fmt.pre = false;

	// count leading spaces
	int		spaces = 0;

	const char* lp = line;
	while (*lp == ' ')
	{
		lp++;
		spaces++;
	}

	if (spaces > 3)
	{
		// too many spaces for formatted text
		fmt.pre = true;
		// preserve indentation
		while (--spaces)
		{
			fputc(' ', fd);
		}
	}
	else if (spaces)
	{
		// indented
		fmt.pre = true;

		if ((lp[0] == '*') && (lp[1] == ' '))
		{
			lp += 2;
			fmt.indent = spaces;
			memset(fmt.indents, 'u', spaces);
			fmt.pre = false;
		}
		else if (isdigit(lp[0]))
		{
			const char*	ep = lp;
			while (isdigit(*ep))	ep++;
			if ((ep[0] == '.') && (ep[1] == ' '))
			{
				lp = ep + 2;
				fmt.indent = spaces;
				memset(fmt.indents, 'o', spaces);
				fmt.pre = false;
			}
		}

		// preserve indentation
		if (fmt.pre)
		{
			while (--spaces)
			{
				fputc(' ', fd);
			}
		}
	}
	else
	{
		if (lp[0] == '[')
		{
			const char*	lp2 = lp + 1;
			char*	num = GetNumber(lp2);
			if (*lp2 == ']')
			{
				lp = lp2 + 1;
				fprintf(fd, "<a name=\"%s\">[%s]", num, num);
			}
		}
		else if (lp[0] == '>')
		{
			do
			{
				lp++;
				fmt.quote++;
				while (*lp == ' ')	lp++;
			} while (*lp == '>');
		}
	}

	// send HTML commands
	_fmt.SetFmt(fmt, fd);

	return lp;
}

// WriteAsHTML
//
// write out a page as HTML

bool PageSet::WriteAsHTML(Page* page, FILE* outfd) const
{
	LoadPage(page);
	
	const char*	title = page->GetTitle();
	char*		stitle = MakeSpacedTitle(title);

	fprintf(outfd, "<html><head>%s<title>%s</title></head>\n", METATAG, stitle);
	fprintf(outfd, "<body><h1><a href=\"StartPage\"><img src=\"%s\"></a> ", gIconName);
	fprintf(outfd, "<a href=\"ReverseLink?page=%s\">%s</a></h1>", title, stitle);

	delete[] stitle;

	fprintf(outfd, "<form>\n");

	char*	ls = page->mASCII;

	ParaFmt	_fmt;

	_fmt.indent = 0;
	_fmt.quote = 0;
	_fmt.pre = false;

	while (ls && *ls)
	{
		// get next line
		char*	line = GetLine(ls);

		// set atributes for line
		const char* lp = SetLineAttributes(line, _fmt, outfd);

		// blank?
		if (!line[0] || (_fmt.quote && !lp[0]))
		{
			fprintf(outfd, "<p>\n");
			delete[] line;
			continue;
		}

		if (!_fmt.indent && !_fmt.quote)
		{
			// ----?
			const char*	lp2 = lp;
			while (*lp2 == '-')	lp2++;
			if (!*lp2 && (lp2 - lp >= 4))
			{
				fprintf(outfd, "<hr>\n");
				delete[] line;
				continue;
			}
		}

		if (_fmt.indent)	fprintf(outfd, "<li>");

		// write out - italic set if quote level is odd
		WriteLine(lp, outfd, (_fmt.quote & 1) ? true : false);

		if (_fmt.indent)	fprintf(outfd, "</li><p>");

		fputc('\n', outfd);

		// loop
		delete[] line;
	}

	ParaFmt	nfmt;

	nfmt.indent = 0;
	nfmt.quote = 0;
	nfmt.pre = false;
	
	_fmt.SetFmt(nfmt, outfd);

	fprintf(outfd, "</form>\n<hr><p>\n");
	fprintf(outfd, "<a href=\"EditPage?page=%s\">EditPage</a> (last edited %s)<br>\n", title, page->GetTimeString());
	fprintf(outfd, "<a href=\"FindPage\">FindPage</a><br>\n");
	fprintf(outfd, "</body></html>\n");

	UnloadPage(page);

	return true;
}
