// WikiServer
//
// Wiki server class

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "..\server.h"

#include "page.h"
#include "wikiserver.h"

#include "..\utility.h"

WikiServer::WikiServer(int port, FILE* logfd, const char* rootdir) : Server(port, logfd)
{
	mPageSet = new PageSet(rootdir);

	// create meta-pages
	mEditPage = mPageSet->CreatePage("EditPage", PF_META);
	mDumpAllPages = mPageSet->CreatePage("DumpAllPages", PF_META);
	mListAllPages = mPageSet->CreatePage("ListAllPages", PF_META);
	mRecentChanges = mPageSet->CreatePage("RecentChanges", PF_META);
	mReverseLink = mPageSet->CreatePage("ReverseLink", PF_META);
}

WikiServer::~WikiServer()
{
	delete mPageSet;
}

const char* WikiServer::GetServerName()
{
	return "Eddie's WikiServer V0.95";
}

int WikiServer::Get(TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	// deal with the IE5 bug ... POSTs sent as GETs
	const Header*	titlep = rq->GetURIData()->Find("title_search");

	if (titlep)
	{
		if (!rq->ConvertGetToPost())	return 500;
		return TitleSearch(titlep->mData, file, rq, hdrs);
	}

	const Header*	fullp = rq->GetURIData()->Find("full_search");

	if (fullp)
	{
		if (!rq->ConvertGetToPost())	return 500;
		return FullSearch(fullp->mData, file, rq, hdrs);
	}

	// get URI; StartPage is the default
	const char*	uri = rq->GetURI();
	if (uri[0] == '/')	uri++;
	if (!uri[0])	uri = "StartPage";

	// look for a .
	const char*	uptr = uri;
	while (*uptr && (*uptr != '.'))	uptr++;
	if (*uptr)
	{
		return GetRaw(file, uri, uptr + 1, hdrs);
	}

//K+
	char * duri = new char[strlen(uri)+1];
	strcpy(duri, uri);
//	Utf8ToAscii(duri);
	URLDecode(duri);

	// find it
//	Page*	pp = mPageSet->FindPage(uri);
	Page*	pp = mPageSet->FindPage(duri);

	if (pp)
	{
//K+
		free(duri);
		if ((pp->GetFlags() & PF_META) && rq->GetURIData())
		{
			return Cmd(pp, file, rq, hdrs);
		}

		// open file
		FILE*	fd = file.Open("w");
		if (!fd)	return 500;

		hdrs.AddHeader("Content-Type", "text/html");
		return mPageSet->WriteAsHTML(pp, fd) ? 200 : 500;
	}


//K	if (PageSet::IsValidTitle(uri))
//K	{
		// forge an EditPage?page=<page>
//		rq->GetURIData()->AddHeader("page", uri);
		rq->GetURIData()->AddHeader("page", duri);
		hdrs.AddHeader("Content-Type", "text/html");
//K+
		free(duri);

		return EditPage(file, rq, hdrs);
//K	}

//K	return 404;
}

int WikiServer::GetRaw(TempFile& file, const char* filename, const char* ext, Headers& hdrs)
{
	const char*	pname = mPageSet->GetPathName();

	// disallow filenames with \ in them for security
	if (strchr(filename, '\\'))	return 403;

	// make filename
	char*	fname = new char[strlen(pname) + strlen(filename) + 2];
	strcpy(fname, pname);
	strcat(fname, "\\");
	strcat(fname, filename);

	// open
	FILE*	ifd = fopen(fname, "rb");
	delete[] fname;
	if (!ifd)	return 404;

	// get length
	fseek(ifd, 0, SEEK_END);
	size_t	len = ftell(ifd);
	fseek(ifd, 0, SEEK_SET);

	// read data
	char*	dbuf = new char[len];
	fread(dbuf, 1, len, ifd);
	fclose(ifd);

	// pipe over
	FILE*	fd = file.Open("wb");
	if (fd)	fwrite(dbuf, 1, len, fd);

	delete[] dbuf;

	if (!fd)	return 500;

	// add correct Content-Type
	if (!stricmp(ext, "gif"))
	{
		hdrs.AddHeader("Content-Type", "image/gif");
	}
	else if (!stricmp(ext, "jpeg") || !stricmp(ext, "jpg"))
	{
		hdrs.AddHeader("Content-Type", "image/jpeg");
	}
	else if (!stricmp(ext, "html") || !stricmp(ext, "htm"))
	{
		hdrs.AddHeader("Content-Type", "text/html");
	}
	else if (!stricmp(ext, "txt"))
	{
		hdrs.AddHeader("Content-Type", "text/plain");
	}
	else
	{
		hdrs.AddHeader("Content-Type", "application/unknown");
	}

	return 200;
}

int WikiServer::Post(TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	// always HTML
	hdrs.AddHeader("Content-Type", "text/html");

	// check URI
	const char* uri = rq->GetURI();
	if (uri[0] == '/')	uri++;
	if (uri[0])	return 404;

	// look for text
	const Header*	pagep = rq->GetPostData()->Find("page");

	if (pagep)
	{
		return SubmitPage(pagep->mData, file, rq, hdrs);
	}

	// look for title_search
	const Header*	titlep = rq->GetPostData()->Find("title_search");

	if (titlep)
	{
		return TitleSearch(titlep->mData, file, rq, hdrs);
	}

	// look for full_search
	const Header*	fullp = rq->GetPostData()->Find("full_search");

	if (fullp)
	{
		return FullSearch(fullp->mData, file, rq, hdrs);
	}

	return 400;
}

int WikiServer::SubmitPage(const char* pagename, TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	// get other parameters
	const	Header*	textp = rq->GetPostData()->Find("text");
	if (!textp)	return 400;

	const	Header*	seqnop = rq->GetPostData()->Find("seqno");

	// define the page
	Page*	pp = mPageSet->CreatePage(pagename);
	if (!pp)	return 500;

	if (pp->GetFlags() & PF_META)	return 400;

	if (seqnop)
	{
		int	seqno = atoi(seqnop->mData);

		if (seqno < pp->GetSeqNo())
		{
			// earlier version - show merge page
			FILE*	fd = file.Open("w");
			if (!fd)	return 500;

			return mPageSet->WriteMergePage(pp, textp->mData, fd) ? 200 : 500;
		}
	}

	mPageSet->EditPage(pp, textp->mData, strlen(textp->mData));

	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	return mPageSet->WriteEditedMessage(pp, fd) ? 200 : 500;
}

int WikiServer::TitleSearch(const char* string, TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	mPageSet->TitleSearch(string, fd);

	return 200;
}

int WikiServer::FullSearch(const char* string, TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	mPageSet->FullSearch(string, fd);

	return 200;
}

int WikiServer::Cmd(Page* page, TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	if (page == mEditPage)		return EditPage(file, rq, hdrs);
	if (page == mDumpAllPages)	return DumpAllPages(file, rq, hdrs);
	if (page == mListAllPages)	return ListAllPages(file, rq, hdrs);
	if (page == mRecentChanges)	return RecentChanges(file, rq, hdrs);
	if (page == mReverseLink)	return ReverseLink(file, rq, hdrs);

	hdrs.AddHeader("Content-Type", "text/html");

	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	if (page)
	{
		fprintf(fd, "<b>%s</b><p>\n", page->GetTitle());
	}
	else
	{
		fprintf(fd, "<b>(No page)</b><p>\n");
	}

	fprintf(fd, "<li>\n");
	rq->GetURIData()->Write(fd);
	fprintf(fd, "</li>\n");

	return 200;
}

int WikiServer::EditPage(TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	const Header*	value = rq->GetURIData()->Find("page");
	if (!value)	return 400;

	const Header*	rows = rq->GetURIData()->Find("rows");
	const Header*	cols = rq->GetURIData()->Find("cols");

	int	nrows = rows ? atoi(rows->mData) : 0;
	int	ncols = cols ? atoi(cols->mData) : 0;

	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	// output page
	hdrs.AddHeader("Content-Type", "text/html");
	return mPageSet->WriteEditPage(value->mData, nrows, ncols, fd) ? 200 : 400;
}

int WikiServer::DumpAllPages(TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	FILE*	fd = file.Open("wb");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", "application/wikiserver");
	mPageSet->DumpAllPages(fd);

	return 200;
}

int WikiServer::ListAllPages(TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", "text/html");
	mPageSet->ListAllPages(fd);

	return 200;
}

int WikiServer::RecentChanges(TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", "text/html");
	mPageSet->RecentChanges(fd);

	return 200;
}

int WikiServer::ReverseLink(TempFile& file, HTTP_RQ* rq, Headers& hdrs)
{
	const Header*	value = rq->GetURIData()->Find("page");
	if (!value)	return 400;

	Page*	page = mPageSet->FindPage(value->mData);
	if (!page)	return 404;

	FILE*	fd = file.Open("w");
	if (!fd)	return 500;

	hdrs.AddHeader("Content-Type", "text/html");
	mPageSet->ReverseLink(page, fd);

	return 200;
}
