// server.cpp
//
// HTTP server

#include <stdio.h>
#include <stdlib.h>

#include "server.h"

Server::Server(int port, FILE* logfd) : mPort(port), mLog(logfd)
{
}

Server::~Server()
{
}

void Server::Log(char* fmt, ...)
{
	if (!mLog)	return;

	static char	buffer[4096];

	va_list	args;
	va_start(args, fmt);
	vsprintf(buffer, fmt, args);
	va_end(args);

	fputs(buffer, mLog);
	fflush(mLog);
}

void Server::AddServerHeaders(Headers& hdrs)
{
	char	buffer[1024];

	sprintf(buffer, "Eddie's MiniWebServer V0.9 running %s\n", GetServerName());

	hdrs.AddHeader("Server", buffer);
}

void Server::Run()
{
	// init server files
	CreateErrorPages();

	// begin Windows sockets
	WORD	wVersionRequested = MAKEWORD(2, 0);
	WSADATA	wsaData;

	if (WSAStartup(wVersionRequested, &wsaData))
	{
		Log("FATAL: Cannot initialize WinSock\n");
		return;
	}

	// get socket
	SOCKET	s;
	int		rc = GetPort(s);

	if (rc == -1)
	{
		Log("FATAL: Cannot open socket\n");
		return;
	}
	else if (rc == -2)
	{
		Log("FATAL: Address in use\n");
		return;
	}

	Log("Server init OK\n"
		"--------------\n");

	// serve
	for (;;)
	{
		struct sockaddr raddr;
		struct sockaddr_in *rap = (struct sockaddr_in *)&raddr;

		int	raddrlen = sizeof(raddr);

		SOCKET as;

		// accept connection
		for (;;)
		{
			as = accept(s, &raddr, &raddrlen);
			if (as != INVALID_SOCKET)					break;		// connected
			if (WSAGetLastError() == WSAEINTR)			continue;	// harmless - try again
			if (WSAGetLastError() != WSAEWOULDBLOCK)	break;		// actual error
			Sleep(0);
		}

		if (as == INVALID_SOCKET)
		{
			Log("Accept failed\n");
		}
		else
		{
			Log("Accept: %s:%d\n", inet_ntoa(rap->sin_addr), mPort);

			// set new socket to non-blocking
			u_long	one = 1;
			if (ioctlsocket(as, FIONBIO, &one) != SOCKET_ERROR)
			{
				// read message from socket
				HTTP_RQ* rq = ReadMessage(as);

				if (rq)
				{
					if (!rq->GetReq())
					{
						Log("Bad HTTP command\n");
						HTTPError(as, 400);
					}
					else
					{
						RunCmd(as, rq);
					}

					delete rq;
				}
				else
				{
					Log("Receive error\n");
					HTTPError(as, 500);
				}
			}

			closesocket(as);
		}
		Log("-----\n");
	}
}

int Server::GetPort(SOCKET& s)
{
	// create socket
	struct sockaddr_in inaddr;

	memset(&inaddr, 0, sizeof(inaddr));

	inaddr.sin_family		= AF_INET;
	inaddr.sin_addr.s_addr	= INADDR_ANY;
	inaddr.sin_port			= htons(mPort);

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s == INVALID_SOCKET)	return -1;

	// bind socket
	if (bind(s, (struct sockaddr *)&inaddr, sizeof(inaddr)) == SOCKET_ERROR)
	{
		closesocket(s);
		if (WSAGetLastError() == WSAEADDRINUSE)	return -2;
		else									return -1;
	}

	// listen
	while (listen(s, 5) == SOCKET_ERROR)
	{
		if (WSAGetLastError() != WSAEINTR)
		{
			closesocket(s);
			return -1;
		}
	}

	// set to non-blocking mode (should be the default)
	u_long zero = 0;
	ioctlsocket(s, FIONBIO, &zero);

	return 0;
}

static int Send(SOCKET s, const char* dbuf, int dlen, int flags)
{
	int olen = dlen;
	int	slen;

	for (;;)
	{
		if (dlen < 65536)
		{
			slen = send(s, dbuf, dlen, flags);
			if (slen != SOCKET_ERROR)
			{
				if (slen == dlen)	return olen;
				dbuf += slen;
				dlen -= slen;
			}
			else
			{
				int	err = WSAGetLastError();

				if (err != WSAEWOULDBLOCK)	break;	// proper error
			}
			Sleep(0);
		}
		else
		{
			slen = send(s, dbuf, 65536, flags);
			if (slen != SOCKET_ERROR)
			{
				dbuf += 65536;
				dlen -= 65536;
			}
			else
			{
				int	err = WSAGetLastError();

				if (err != WSAEWOULDBLOCK)	break;	// proper error
			}
			Sleep(0);
		}
	}

	return SOCKET_ERROR;
}

static int Recv(SOCKET s, char* buf, int dlen, int flags)
{
	int	rlen;

	for (;;)
	{
		rlen = recv(s, buf, 512, 0);
		if (!rlen)									return SOCKET_ERROR;	// connection is closed
		if (rlen != SOCKET_ERROR)					return rlen;
		if (WSAGetLastError() != WSAEWOULDBLOCK)	break;					// real error
		Sleep(0);
	}

	return SOCKET_ERROR;
}

// CreateErrorPages
//
// create pages used by HTTPError function

void Server::CreateErrorPages()
{
	CreateErrorPage(400, "Bad request", "The server cannot complete your request");
	CreateErrorPage(401, "Unauthorized", "You are not authorized to view that page");
	CreateErrorPage(403, "Forbidden", "You are forbidden to view that page");
	CreateErrorPage(404, "Not found", "The page you requested was not found");

	CreateErrorPage(500, "Internal error", "The server has suffered an internal error");
	CreateErrorPage(501, "Not implemented", "The server does not implement that request");
}

void Server::CreateErrorPage(int status, char* desc, char* verbiage)
{
	char	temp[MAX_PATH];
	sprintf(temp, "Error%d.html", status);
	FILE*	fd = fopen(temp, "w");
	if (fd)
	{
		fprintf(fd, "%s\n", StatusLine(status));

		Headers	hdr;

		hdr.AddHeader("Content-Type", "text/html");
		AddServerHeaders(hdr);

		hdr.Write(fd);

		fprintf(fd, "\n<html><head><title>\n%d - %s\n", status, desc);
		fprintf(fd, "</title></head>\n<body><h1>\n%d - %s\n</h1><p>\n", status, desc);
		fprintf(fd, "%s\n</body></html>\n", verbiage);

		fclose(fd);
	}
}

void Server::HTTPError(SOCKET s, int status)
{
	const char* err = StatusLine(status);
	Log("%s\n", err);

	// look for response file
	char temp[MAX_PATH];
	sprintf(temp, "Error%d.html", status);
	FILE*	fd = fopen(temp, "rb");
	if (fd)
	{
		fseek(fd, 0, SEEK_END);
		size_t dlen = ftell(fd);
		fseek(fd, 0, SEEK_SET);

		char*	dbuf = new char[dlen];
		fread(dbuf, 1, dlen, fd);
		fclose(fd);

		Send(s, dbuf, dlen, 0);
		delete[] dbuf;
	}
	else
	{
		sprintf(temp, "%s\x0D\0xA\x0D\x0A", err);
		Send(s, temp, strlen(temp), 0);
	}
}

// http_terminator
//
// looks for [<0D>]<0A>[<0D>]<0A> in a block
//
// returns position of next character, else 0

static int http_terminator(char* mbuf, int len)
{
	int	conseq = 0;

	for (int ii = 0; ii < len; ii++)
	{
		if (mbuf[ii] == 0x0A)
		{
			conseq++;
			if (conseq == 2)	return (ii + 1);
		}
		else if (mbuf[ii] == 0x0D)
		{
			// ignore
		}
		else
		{
			conseq = 0;
		}
	}

	return 0;
}

// ReadMessage
//
// read a message

HTTP_RQ* Server::ReadMessage(SOCKET s)
{
	char	buf[512];		// buffer for data
	char*	mbuf = NULL;	// cumulative buffer
	size_t	len = 0;		// length of mbuf[]
	size_t	tpos = 0;		// end of header

	for (;;)
	{
		int	rlen = Recv(s, buf, 512, 0);

		if (rlen == SOCKET_ERROR)
		{
			delete[] mbuf;
			return NULL;
		}

		if (len)
		{
			// re-alloc
			char*	nbuf = new char[len + rlen + 1];
			memcpy(nbuf, mbuf, len);
			memcpy(nbuf + len, buf, rlen);
			delete[] mbuf;
			mbuf = nbuf;
			len += rlen;

			if (tpos = http_terminator(mbuf, len))	break;
		}
		else
		{
			mbuf = new char[rlen + 1];
			memcpy(mbuf, buf, rlen);
			len = rlen;

			if (tpos = http_terminator(mbuf, len))	break;
		}
	}

	// copy header out
	char*	hdr = new char[tpos + 1];
	memcpy(hdr, mbuf, tpos);
	hdr[tpos] = '\0';

	char*	post = NULL;
	if (len > tpos)
	{
		post = new char[len - tpos];
		memcpy(post, mbuf + tpos, len - tpos);
	}
	delete[] mbuf;
	mbuf = post;
	len -= tpos;

	// decode header
	HTTP_RQ*	rq = new HTTP_RQ;
	rq->AttachHeaders(hdr);

	Log("%s %s %s\n", rq->GetReq(), rq->GetURI(), rq->GetHTTPVer());

	// get content length
	const char*	cl = rq->GetHeader("Content-Length");
	if (cl)
	{
		size_t datalen = atoi(cl);

		while (datalen > len)
		{
			int	rlen = Recv(s, buf, 512, 0);

			if (rlen == SOCKET_ERROR)
			{
				delete[] mbuf;
				delete rq;
				return NULL;
			}

			if (len)
			{
				// re-alloc
				char*	nbuf = new char[len + rlen + 1];
				memcpy(nbuf, mbuf, len);
				memcpy(nbuf + len, buf, rlen);
				delete[] mbuf;
				mbuf = nbuf;
				len += rlen;
			}
			else
			{
				mbuf = new char[rlen + 1];
				memcpy(mbuf, buf, rlen);
				len = rlen;
			}
		}
	}

	// attach data buffer
	rq->AttachData(mbuf, len);

	return rq;
}

void Server::RunCmd(SOCKET s, HTTP_RQ* rq)
{
	// create temp file
	TempFile	header("HDR");
	TempFile	body("BDY");

	// run command
	Headers		hdrs;
	int			status = 501;	// not implemented
	bool		senddata = true;

	if (!stricmp(rq->GetReq(), "get"))
	{
		status = Get(body, rq, hdrs);
	}
	if (!stricmp(rq->GetReq(), "head"))
	{
		status = Get(body, rq, hdrs);
		senddata = false;
	}
	else if (!stricmp(rq->GetReq(), "post"))
	{
		status = Post(body, rq, hdrs);
	}

	// close file
	body.Close();

	if (status != 200)
	{
		HTTPError(s, status);
		return;
	}

	// open data for reading
	FILE*	bfd = body.Open("rb");
	if (!bfd)
	{
		Log("Can't reopen temp body file\n");
		HTTPError(s, 500);
		return;
	}

	// get size
	fseek(bfd, 0, SEEK_END);
	size_t	blen = ftell(bfd);
	fseek(bfd, 0, SEEK_SET);

	// add Content-Length header
	char	temp[16];
	sprintf(temp, "%d", blen);
	hdrs.AddHeader("Content-Length", temp);

	// write headers
	FILE*	hfd = header.Open("w");
	if (!hfd)
	{
		Log("Can't open temp head file\n");
		HTTPError(s, 500);
		return;
	}

	Log("%s\n", StatusLine(status));
	fprintf(hfd, "%s\n", StatusLine(status));
	hdrs.Write(hfd);
	fprintf(hfd, "\n");
	fclose(hfd);

	// open headers for reading
	hfd = header.Open("rb");
	if (!hfd)
	{
		Log("Can't reopen temp head file\n");
		HTTPError(s, 500);
		return;
	}

	// get size
	fseek(hfd, 0, SEEK_END);
	size_t hlen = ftell(hfd);
	fseek(hfd, 0, SEEK_SET);

	// load both files in
	char*	data = new char[hlen + blen];
	fread(data, 1, hlen, hfd);
	fread(data + hlen, 1, blen, bfd);
	
	fclose(hfd);
	fclose(bfd);

	// send the data
	if (senddata)
	{
		Send(s, data, hlen, 0);
		Send(s, data + hlen, blen, 0);
	}
	else
	{
		Send(s, data, hlen, 0);
	}
	
	delete[] data;
}
